@@ -221,14 +221,17 @@ namespace config {
 				"https://reseed-pl.i2pd.xyz/,"
 				"https://www2.mk16.de/,"
 			    "https://i2p.ghativega.in/,"
-			    "https://i2p.novg.net/"
+			    "https://i2p.novg.net/,"
+			    "https://reseed.is.prestium.org/,"
+				"https://reseed.us.prestium.org/"
 			),                                                            "Reseed URLs, separated by comma")
 			("reseed.yggurls", value<std::string>()->default_value(
 				"http://[324:71e:281a:9ed3::ace]:7070/,"
 				"http://[301:65b9:c7cd:9a36::1]:18801/,"
 				"http://[320:8936:ec1a:31f1::216]/,"
 				"http://[306:3834:97b9:a00a::1]/,"
-				"http://[316:f9e0:f22e:a74f::216]/"
+				"http://[316:f9e0:f22e:a74f::216]/,"
+			    "http://[300:eaff:7fab:181b::e621]:7170"
 			),                                                            "Reseed URLs through the Yggdrasil, separated by comma")
 		;
 
